#include <AccelStepper.h>
#include <MultiStepper.h>


//Neck Stepper
AccelStepper Neck(1,40,41);
int neckCenteringDist = 250;
int neckSpeed = 2000;
int neckSensor;
int neckPin = 22;
bool isCentered = false;


//Drive Steppers
AccelStepper RightFrontWheel(1, 20, 21); // Stepper1
AccelStepper RightBackWheel(1, 44, 45);   // Stepper2
AccelStepper LeftFrontWheel(1, 18, 19);  // Stepper3
AccelStepper LeftBackWheel(1, 42, 43);  // Stepper4
MultiStepper steppers;
long positions[4] = {0,0,0,0};
int MovementCommand [4] = {0,0,0,0};
int wheelSpeed = 2000;

float angleDistComp = 0.4;
float DistToSteps = 0.557;

int Angle = 0;
int Distance = 0;


//BrushStepper
AccelStepper Brush(AccelStepper::FULL4WIRE,26,28,27,29);
int brushUP = 500;
int brushDOWN = 0;
int brushHeight = 0;

//Sensors
int ESpin_fl = 37;
int ESpin_fr = 35;
int ESpin_bl = 36;
int ESpin_br = 34;

int ES_fl; 
int ES_fr;
int ES_bl;
int ES_br;

bool edgeDetected = false;

//int IRpin = 25;

//Other
bool Run = true;

void setup() 
{
  //Neck Stepper
  Neck.setMaxSpeed(200);
  Neck.setAcceleration(10000); 
  pinMode(neckPin,INPUT);

  //Brush Stepper
  Brush.setMaxSpeed(500);
  Brush.setAcceleration(10000);

  //Drive Steppers
  LeftFrontWheel.setMaxSpeed(wheelSpeed);
  RightFrontWheel.setMaxSpeed(wheelSpeed);
  RightBackWheel.setMaxSpeed(wheelSpeed);
  LeftBackWheel.setMaxSpeed(wheelSpeed);
  steppers.addStepper(LeftFrontWheel);
  steppers.addStepper(LeftBackWheel);
  steppers.addStepper(RightFrontWheel);
  steppers.addStepper(RightBackWheel);

  //Sensors
  pinMode(ESpin_fl,INPUT);
  pinMode(ESpin_fr,INPUT);
  pinMode(ESpin_bl,INPUT);
  pinMode(ESpin_br,INPUT);
  
  //Other
  Serial.begin(115200);

  delay(5000);
}
 
void loop() 
{

//  HomeNeck();

int DrawList [126] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0};
int AngleList [126] = {343,330,316,303,289,276,263,248,231,215,200,186,180,174,160,145,129,112,97,84,71,57,44,31,18,118,85,73,61,51,42,33,20,6,0,355,341,328,318,309,299,287,275,270,180,90,90,85,73,61,51,42,33,20,6,0,355,341,328,318,309,299,287,275,270,180,90,90,0,90,189,270,105,36,28,20,12,4,356,348,340,332,325,317,309,301,293,285,277,269,261,253,245,237,229,221,213,205,335,327,319,311,303,295,287,279,271,263,255,247,239,231,223,215,208,200,192,184,176,168,160,152,144,136,44,105};
int LengthList [126] = {12,12,12,12,12,12,15,14,13,14,14,16,41,16,14,14,13,14,15,12,12,12,12,12,12,87,17,15,7,7,6,7,14,16,42,16,14,7,6,7,7,15,17,47,150,47,138,17,15,7,7,6,7,14,16,42,16,14,7,6,7,7,15,17,47,150,47,91,150,94,71,83,192,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,28,503,503,192};

  if (Run == true)
  {
    int counter = 0;
    while (counter < 126)
    {
      if (DrawList[counter] == 1)
      {
        brushHeight = brushDOWN;
      }
      else
      {
        brushHeight = brushUP;
      }
      LinearMove(AngleList[counter],LengthList[counter],brushHeight);
      counter++;
    }
   LinearMove(90,400,brushUP);
  }
  
  Run = false;
}

///////////////////////////////////////////////////////////////////

void CheckEdgeSensors()
{
  ES_fl = digitalRead(ESpin_fl);
  ES_fr = digitalRead(ESpin_fr);
  ES_bl = digitalRead(ESpin_bl);
  ES_br = digitalRead(ESpin_br);
  if (ES_fl or ES_fr or ES_bl or ES_br)
  {
    edgeDetected = true;
  }
}

///////////////////////////////////////////////////////////////////

void SquareMove(int InAngle,int InDistance,int InBrushHeight)
{
  int angleList [4] = {InAngle,InAngle+90,InAngle+180,InAngle+270};
  int i = 0;
  while (i < 4 and edgeDetected == false)
  {
      int brushChance = random(0,100);
    if (brushChance >= 20)
    {
      brushHeight = brushDOWN;
    }
    else
    {
      brushHeight = brushUP;
    }
    LinearMove(angleList[i],InDistance,brushHeight);
    i++;
  }
}

///////////////////////////////////////////////////////////////////

void MoveSteppers(int Command[4]) 
{
  positions[0] += Command[0];
  positions[1] += Command[1];
  positions[2] += Command[2];
  positions[3] += Command[3];
  steppers.moveTo(positions);
  while (steppers.run() == 1 and edgeDetected == false)
  {
    CheckEdgeSensors();
  }
}


///////////////////////////////////////////////////////////////////

void LinearMove(int InAngle,int InDistance,int InBrushHeight)
{
  int EvalAngle = abs(InAngle%360);
  int AngleMod = EvalAngle%90;

//Move Neck to Direction
//  int neckTarget = EvalAngle;
//  if (neckTarget <= 180)
//  {
//    MoveNeck(neckTarget);
//  }
//  else
//  {
//    neckTarget = 360 - neckTarget;
//    MoveNeck(-neckTarget);
//  }

//Eval Angle to modify wheel directions and speeds
  if (0 <=EvalAngle && EvalAngle < 90)
  {
    MovementCommand[0] = -45;
    MovementCommand[1] = AngleMod - 45;
    MovementCommand[2] = -AngleMod + 45;
    MovementCommand[3] = 45;
  }
  else if (EvalAngle < 180)
  {
    MovementCommand[0] = AngleMod - 45;
    MovementCommand[1] = 45;
    MovementCommand[2] = -45;
    MovementCommand[3] = -AngleMod + 45;
  }
  else if (EvalAngle < 270)
  {
    MovementCommand[0] = 45;
    MovementCommand[1] = -AngleMod + 45;
    MovementCommand[2] = AngleMod - 45;
    MovementCommand[3] = -45;
  }
  else
  {
    MovementCommand[0] = -AngleMod + 45;
    MovementCommand[1] = -45;
    MovementCommand[2] = 45;
    MovementCommand[3] = AngleMod - 45;
  }
  
  //Modify CounterLimit to distance
  if (AngleMod > 45)
  {
    AngleMod = 90 - AngleMod;
  }
  //Convert Dist in mm to number of 45 step cycles
  float tempDistFactor = InDistance*DistToSteps; // old InDistance/1.795; older InDistance/1.75
  int distFactor = tempDistFactor;
  //Compensate number of cycles for Direction 
  float tempAngleFactor = distFactor * ((sqrt(sq(45)-sq(AngleMod-45))/45)*angleDistComp);
  //float tempAngleFactor = distFactor * (tan(AngleMod)*angleDistComp);
  int angleFactor = tempAngleFactor;
  //Compute Final number of cycles
  int numCycles = distFactor + angleFactor;
  // Multiply the 45 step moves by number of cycles
  MovementCommand[0] *= numCycles;
  MovementCommand[1] *= numCycles;
  MovementCommand[2] *= numCycles;
  MovementCommand[3] *= numCycles;
  
  //Move Pen
  MoveBrush(InBrushHeight);

  //Exicute Move
  MoveSteppers(MovementCommand);
}

///////////////////////////////////////////////////////////////////

void HomeNeck()
{
  if (isCentered == false)
  {
    Neck.setMaxSpeed(200);
    neckSensor = digitalRead(neckPin);
    if (neckSensor == 1)
    {
      isCentered = true;
    }
    else
    {
      Neck.moveTo(neckCenteringDist);
      while (Neck.run() == 1)
      {
        neckSensor = digitalRead(neckPin);
        if (neckSensor == 1)
        {
          isCentered = true;
          break;
        }
      }
      if (isCentered != true);
      {
        Neck.moveTo(-neckCenteringDist);
        while (Neck.run() == 1)
        {
          neckSensor = digitalRead(neckPin);
          if (neckSensor == 1)
          {
            isCentered = true;
            break;
          }
        }
      } 
    }
    if (isCentered)
    {
      Neck.setCurrentPosition(0);
      Serial.println("Neck Centered");
    }
    else
    {
      Serial.println("Neck Center Not Found");
    }
  }
  Neck.setMaxSpeed(neckSpeed);
}

///////////////////////////////////////////////////////////////////

void MoveNeck(float targetAngle)
{
  float tempTarget = (targetAngle/360)*3200;
  int target = tempTarget;
  Neck.moveTo(target);
  while (Neck.run() == 1){}
}

///////////////////////////////////////////////////////////////////

void MoveBrush(int target)
{
  Brush.moveTo(target);
  while (Brush.run() == 1){}
}
